# This code is part of Tergite
#
# (C) Copyright Miroslav Dobsicek 2021
# (C) Copyright Axel Andersson 2022
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from dataclasses import asdict, dataclass, field


@dataclass
class ProviderAccount:
    service_name: str
    url: str
    token: str = None
    extras: dict = field(default_factory=dict)

    def to_dict(self: object) -> dict:
        return asdict(self)



# This code is part of Tergite
#
# (C) Copyright Axel Andersson 2022
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# This code was refactored from the original on 22nd September, 2023 by Martin Ahindura
"""Handles the serialization of job objects"""
import json
from itertools import groupby
from typing import Any, List, Tuple, Union  # , Dict

from qiskit.circuit.parameterexpression import ParameterExpression


def iqx_rle(seq: List[Any]) -> List[Union[Tuple[Any], Tuple[Any, int]]]:
    """Run-length encodes a sequence.

    Constant subsequences are stored as a single term and count, rather than as the original subsequence.
    Counts equal to 1 are omitted.

    Args:
        seq: the sequence to encode

    Returns:
        A list of tuples that represent the encoding for the sequence
    """
    seq = [(k, sum(1 for _ in g)) for k, g in groupby(seq)]
    return [(c, rep) if rep > 1 else (c,) for c, rep in seq]


# The below code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# This code is part of Tergite
#
# (C) Copyright Axel Andersson 2022
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# ---------------- ALTERATION NOTICE ---------------- #
# This code has been derived from the Qiskit json encoder in the IBMQ providers github.
# The reason is to remove IBMQ as a dependency, but still be able to use the functionality.
# It has not been altered.


class IQXJsonEncoder(json.JSONEncoder):
    """A json encoder for qobj"""

    def __encode(self, param: Any) -> Any:
        """
        Convert dictionary to contain only JSON serializable types. For example,
        if the key is a Parameter we convert it to a string.
        """
        if isinstance(param, dict):
            param_bind_str = {}
            for key in param.keys():
                value = self.__encode(param[key])

                if isinstance(key, (bool, float, int, str)) or key is None:
                    param_bind_str[key] = value
                else:
                    param_bind_str[str(key)] = value
            return param_bind_str
        elif isinstance(param, list):
            return [self.__encode(p) for p in param]
        else:
            return param

    def encode(self, o: Any) -> str:
        """
        Return a JSON string representation of a Python data structure.
        """
        new_o = self.__encode(o)
        return super().encode(new_o)

    def default(self, o: Any) -> Any:
        # Convert numpy arrays:
        if hasattr(o, "tolist"):
            return o.tolist()
        # Use Qobj complex json format:
        if isinstance(o, complex):
            return (o.real, o.imag)
        if isinstance(o, ParameterExpression):
            try:
                return float(o)
            except (TypeError, RuntimeError):
                val = complex(o)
                return val.real, val.imag
        return json.JSONEncoder.default(self, o)



# This code is part of Tergite
#
# (C) Copyright Miroslav Dobsicek 2020, 2021
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# This code was refactored from the original on 22nd September, 2023 by Martin Ahindura
"""Handles the loading and saving of configuration to tergiterc file"""
import dataclasses
import pathlib
import re
from configparser import ConfigParser
from typing import List, Optional

from .provider_account import ProviderAccount

TERGITERC_FILE = pathlib.Path.home() / ".qiskit" / "tergiterc"

REST_API_MAP = {
    "jobs": "/jobs",
    "result": "/result",
    "download_url": "/download_url",
    "calibrations": "/calibrations",
    "backends": "/backends",
}


class Tergiterc:
    """the Configuration parser for tergiterc files"""

    def __init__(self, rc_file: pathlib.Path = TERGITERC_FILE):
        """Initializes a Tergiterc instance

        The instance initialized, saves to and retrieves its
        data from file_path

        Args:
            rc_file: the path where the tergiterc file is saved, defaults to ``$HOME/.qiskit/tergiterc``
        """
        self._file_path = rc_file
        self._parser = Tergiterc._get_parser(rc_file)

    def load_accounts(self) -> List["ProviderAccount"]:
        """Retrieves the accounts from the tergiterc file

        Returns:
            list of instances of
                :class:`~tergite.providers.tergite.provider_account.ProviderAccount`
                as read from the tergiterc file
        """
        account_fields = {
            field.name: True for field in dataclasses.fields(ProviderAccount)
        }
        accounts = []

        parser = self._parser
        if not parser:
            return accounts

        sections_all = parser.sections()
        for section in sections_all:
            if section.startswith("service"):
                if not parser.has_option(section, "url"):
                    print(
                        f"Warning: Skipping account provider '{section}'. Invalid configuration."
                    )
                    continue

                section_items = dict(parser.items(section))
                service_name = section.split(" ", 1)[1].strip()

                # extract the extras (since on saving to file, they are flattened)
                account_options = {"extras": {}}
                for k, v in section_items.items():
                    if k in account_fields:
                        account_options[k] = v
                    else:
                        account_options["extras"][k] = v

                new_account = ProviderAccount(service_name, **account_options)
                accounts.append(new_account)

        return accounts

    def save_accounts(self, accounts: List["ProviderAccount"]):
        """Saves the accounts into the tergiterc file

        Args:
            accounts: the list of instances of
                :class:`~tergite.providers.tergite.provider_account.ProviderAccount`
                to save

        Raises:
            Exception: If no accounts are passed
        """
        if not accounts:
            raise Exception("Cannot save account(s). None given.")

        if not self._parser:
            self._parser = ConfigParser()

        for account in accounts:
            # section
            section_name = "service " + account.service_name
            self._parser.add_section(section_name)

            # account configuration
            config = account.to_dict()
            config.update(config.pop("extras"))  # flatten on 'extras'
            config.pop("service_name")  # remove 'service_name'

            for key, value in config.items():
                self._parser.set(section_name, str(key), str(value))

        with self._file_path.open("w") as dest:
            self._parser.write(dest)

    @staticmethod
    def _get_parser(rc_file: pathlib.Path) -> Optional[ConfigParser]:
        """Initializes a :class:`configparser.ConfigParser` instance that
        has read from rc_file

        It returns None if rc_file does not exist.

        Returns:
            Optional[configparser.ConfigParser]: the ConfigParser with the
                rc_file loaded in it if the file exists
        """
        if not rc_file.exists():
            return None

        parser = ConfigParser()
        parser.SECTCRE = re.compile(r"\[ *(?P<header>[^]]+?) *\]")
        parser.read(rc_file)

        return parser



# This code is part of Tergite
#
# (C) Copyright Miroslav Dobsicek 2020, 2021
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# This code was refactored from the original on 22nd September, 2023 by Martin Ahindura
"""Defines the Factory class that contains multiple Tergite Providers."""
import pathlib
from collections import OrderedDict
from typing import TYPE_CHECKING, List, Optional

from .config import TERGITERC_FILE, Tergiterc
from .provider import Provider

if TYPE_CHECKING:
    from .provider_account import ProviderAccount


class Factory:
    """Container of multiple Tergite Provider's, retrievable by service name

    It can be passed a custom path to an rc file: default is `$HOME/.qiskit/tergiterc`
    """

    def __init__(self, rc_file: pathlib.Path = TERGITERC_FILE):
        self._providers = OrderedDict()
        self._tergiterc = Tergiterc(rc_file=rc_file)

        # get a list of provider account objects from tergiterc
        accounts_list = self._tergiterc.load_accounts()
        if not accounts_list:
            print(
                "Warning: No stored provider account found in $HOME/.qiskit/tergiterc"
            )

        # for each item in the accounts list create a Provider object
        for account in accounts_list:
            self._providers[account.service_name] = Provider(account)

    def use_provider_account(
        self, account: "ProviderAccount", save: bool = False
    ) -> Provider:
        """Intializes a new Provider basing on the account passed and returns it.

        Args:
            account: the instance of the
                :class:`~tergite.providers.tergite.provider_account.ProviderAccount`
                for which the provider is to be initialized
            save: whether the account should be persisted to the tergiterc file

        Returns:
            A new instance of the
                :class:`~tergite.providers.tergite.provider.Provider`
                for the given account
        """
        if save:
            self._tergiterc.save_accounts([account])

        new_provider = Provider(account)
        self._providers[account.service_name] = new_provider

        return new_provider

    def providers(self) -> List[str]:
        """Retrieves the service names of providers in this collection

        Returns:
            list of service names of providers in this collection
        """
        return list(self._providers.keys())

    def get_provider(self, service_name=None) -> Optional[Provider]:
        """Retrieves a given provider by service name

        Args:
            service_name: the name by which the given provider is saved

        Returns:
            the provider of the given service name
                if it exists or None if it doesn't
        """
        providers = self._providers

        if not providers:
            raise Exception(
                "No provider account is available. Provide one via Tergite.use_provider_account(..)."
            )

        # get by name or default to the first entry
        return providers.get(service_name, next(iter(providers.values())))



# This code is part of Tergite
#
# (C) Copyright Miroslav Dobsicek 2020, 2021
# (C) Copyright Axel Andersson 2022
# (C) Copyright Martin Ahindura 2023
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# This code was refactored from the original on 22nd September, 2023 by Martin Ahindura
"""Defines the Qiskit provider with which to access the Tergite Quantum Computers"""
import functools
from typing import Dict, List, Optional, Union

import requests
from qiskit.providers import ProviderV1
from qiskit.providers.providerutils import filter_backends

from .backend import OpenPulseBackend, OpenQASMBackend, TergiteBackendConfig
from .config import REST_API_MAP
from .provider_account import ProviderAccount


class Provider(ProviderV1):
    """The Qiskit Provider with which to access the Tergite quantum computers"""

    def __init__(self, /, account: ProviderAccount):
        """Initializes the Provider

        Args:
            account: the instance of the
                :class:tergite.providers.tergite.provider_account.ProviderAccount`
                with which to connect to the Tergite API
        """
        super().__init__()
        self.provider_account = account
        self._malformed_backends = {}

    def backends(
        self, /, name: str = None, filters: callable = None, **kwargs
    ) -> List[Union[OpenPulseBackend, OpenQASMBackend]]:
        """Filters the available backends of this provider.

        Args:
            name: the name of the backend
            filters: a callable to filter the backends with
            kwargs: kwargs to match the available backends with

        Returns:
            A list of instantiated and available OpenPulseBackend, or OpenPulseBackend backends,
                that match the given filter
        """
        available_backends = self.available_backends
        if name in self._malformed_backends:
            exp = {self._malformed_backends[name]}
            raise TypeError(f"malformed backend '{name}', {exp}")

        if name:
            kwargs["backend_name"] = name

        return filter_backends(available_backends.values(), filters=filters, **kwargs)

    @functools.cached_property
    def available_backends(
        self, /
    ) -> Dict[str, Union[OpenPulseBackend, OpenQASMBackend]]:
        """Dictionary of all available backends got from the API"""
        backends = dict()
        backend_configs = self._get_backend_configs()

        for backend_conf in backend_configs:
            if backend_conf.open_pulse:
                obj = OpenPulseBackend(
                    data=backend_conf, provider=self, base_url=self.provider_account.url
                )
            else:
                obj = OpenQASMBackend(
                    data=backend_conf, provider=self, base_url=self.provider_account.url
                )
            backends[obj.name] = obj

        return backends

    def _get_backend_configs(self) -> List[TergiteBackendConfig]:
        """Retrieves the backend configs from which to construct Backend objects"""
        parsed_data = []
        url = f"{self.provider_account.url}{REST_API_MAP['backends']}"

        # reset malformed backends map
        self._malformed_backends.clear()

        headers = self.get_auth_headers()
        response = requests.get(url=url, headers=headers)
        if not response.ok:
            raise RuntimeError(f"GET request for backends timed out. GET {url}")

        records = response.json()
        for record in records:
            try:
                parsed_data.append(TergiteBackendConfig(**record))
            except TypeError as exp:
                self._malformed_backends[record["name"]] = f"{exp}\n{exp.__traceback__}"

        return parsed_data

    def get_auth_headers(self) -> Optional[Dict[str, str]]:
        """Retrieves the auth header for this provider.

        Returns:
            dict of authorization of the authorization headers if account has auth else None
        """
        if self.provider_account.token:
            return {"Authorization": f"Bearer {self.provider_account.token}"}

        return None

    def __str__(self, /):
        return "Tergite: Provider"

    def __repr__(self, /):
        return "<{} from Tergite Qiskit>".format(self.__class__.__name__)



# This code is part of Tergite
#
# (C) Copyright Miroslav Dobsicek 2020, 2021
# (C) Copyright Axel Andersson 2022
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# This code was refactored from the original on 22nd September, 2023 by Martin Ahindura
"""Defines the asynchronous job that executes the experiments."""
import json
from collections import Counter
from pathlib import Path
from tempfile import gettempdir
from typing import TYPE_CHECKING, Optional, Tuple, Union
from uuid import uuid4

import requests
from qiskit.providers import BackendV1, JobV1
from qiskit.providers.jobstatus import JobStatus
from qiskit.qobj import PulseQobj, QasmQobj
from qiskit.result import Result
from qiskit.result.models import ExperimentResult, ExperimentResultData
from requests import Response

from .config import REST_API_MAP
from .serialization import IQXJsonEncoder, iqx_rle

if TYPE_CHECKING:
    from tergite.qiskit.providers.tergite import Provider

    from .backend import TergiteBackend

STATUS_MAP = {
    "REGISTERING": JobStatus.QUEUED,
    "DONE": JobStatus.DONE,
    # TODO: unimplemented status codes
    "INITIALIZING": JobStatus.INITIALIZING,
    "VALIDATING": JobStatus.VALIDATING,
    "RUNNING": JobStatus.RUNNING,
    "CANCELLED": JobStatus.CANCELLED,
    "ERROR": JobStatus.ERROR,
}


class Job(JobV1):
    """A representation of the asynchronous job that handles experiments on a backend"""

    def __init__(self, *, backend: "TergiteBackend", job_id: str, upload_url: str):
        """Initializes the job instance for the given backend

        Args:
            backend: the backed where the job is to run
            job_id: the unique id of the job
            upload_url: URL where the jobs will be uploaded
        """
        super().__init__(backend=backend, job_id=job_id, upload_url=upload_url)
        self.payload: Optional[Union[QasmQobj, PulseQobj]] = None

    def status(self) -> JobStatus:
        response = self._get_job_results()
        if response.ok:
            return STATUS_MAP[response.json()["status"]]
        else:
            raise RuntimeError(f"Failed to GET status of job: {self.job_id()}")

    def submit(self, payload: Union[QasmQobj, PulseQobj], /) -> requests.Response:
        """Submit the job to the backend for execution.

        Args:
            payload: the QasmQobj or PulseQobj object to execute

        Returns:
            requests.Response: the response of the API after submitting the job
        """
        self.metadata["shots"] = payload.config.shots
        self.metadata["qobj_id"] = payload.qobj_id
        self.metadata["num_experiments"] = len(payload.experiments)
        self.payload = payload  # save input data inside job

        job_id = self.job_id()
        job_entry = {
            "job_id": job_id,
            "type": "script",  # ?
        }
        if type(payload) is QasmQobj:
            job_entry["name"] = "qasm_dummy_job"
            job_entry.update(
                {
                    "name": "qasm_dummy_job",
                    "params": {"qobj": payload.to_dict()},
                    "post_processing": "process_qiskit_qasm_runner_qasm_dummy_job",
                }
            )

        elif type(payload) is PulseQobj:
            payload = payload.to_dict()
            # In-place RLE pulse library for compression
            for pulse in payload["config"]["pulse_library"]:
                pulse["samples"] = iqx_rle(pulse["samples"])

            job_entry.update(
                {
                    "name": "pulse_schedule",
                    "params": {"qobj": payload},
                }
            )

        else:
            raise RuntimeError(f"Unprocessable payload type: {type(payload)}")

        # Serialize the job to json
        job_file = Path(gettempdir()) / str(uuid4())
        with job_file.open("w") as dest:
            json.dump(job_entry, dest, cls=IQXJsonEncoder, indent="\t")

        job_upload_url = self.metadata["upload_url"]

        backend: "TergiteBackend" = self.backend()
        provider: "Provider" = backend.provider
        auth_headers = provider.get_auth_headers()

        # Transmit the job POST request
        with job_file.open("r") as src:
            files = {"upload_file": src}
            response = requests.post(job_upload_url, files=files, headers=auth_headers)
            if not response.ok:
                raise RuntimeError(f"Failed to POST job: {job_id}")

        # Delete temporary transmission file
        job_file.unlink()
        return response

    @property
    def download_url(self) -> Optional[str]:
        """The download_url of this job when it is completed"""
        if self.status() != JobStatus.DONE:
            print(f"Job {self.job_id()} has not yet completed.")
            return

        response = self._get_job_results()
        if response.ok:
            return response.json()["download_url"]
        else:
            raise RuntimeError(f"Failed to GET download URL of job: {self.job_id()}")

    @property
    def logfile(self) -> Optional[Path]:
        """The path to the logfile of this job when it is completed"""
        url = self.download_url
        if not url:
            return

        backend: "TergiteBackend" = self.backend()
        provider: "Provider" = backend.provider
        auth_headers = provider.get_auth_headers()

        response = requests.get(url, headers=auth_headers)
        job_id = self.job_id()
        if response.ok:
            job_file = Path(gettempdir()) / (job_id + ".hdf5")
            with open(job_file, "wb") as dest:
                dest.write(response.content)
            return job_file
        else:
            raise RuntimeError(f"Failed to GET logfile of job: {job_id}")

    def cancel(self):
        print("Job.cancel() is not implemented.")
        pass  # TODO: This can be implemented server side with stoppable threads.

    def result(self) -> Optional[Result]:
        """Retrieves the outcome of this job when it is completed.

        It returns None if the job has not yet completed

        Returns:
            Optional[qiskit.result.result.Result]: the outcome of this job
                if it has completed
        """
        if self.status() != JobStatus.DONE:
            print(f"Job {self.job_id()} has not yet completed.")
            return

        response = self._get_job_results()
        if response.ok:
            memory = response.json()["result"]["memory"]
        else:
            raise RuntimeError(f"Failed to GET memory of job: {self.job_id()}")

        # Sanity check
        if not len(memory) == self.metadata["num_experiments"]:
            print(
                "There is a mismatch between number of experiments in the Qobj \
                and number of results recieved!"
            )
        else:
            print("Results OK")

        # Extract results
        experiment_results = list()
        for index, experiment_memory in enumerate(memory):
            experiment_results.append(
                ExperimentResult(
                    header=self.payload.experiments[index].header,
                    shots=self.metadata["shots"],
                    success=True,
                    data=ExperimentResultData(
                        counts=dict(Counter(experiment_memory)),
                        memory=experiment_memory,
                    ),
                )
            )

        backend = self.backend()
        return Result(
            backend_name=backend.name,
            backend_version=backend.backend_version,
            qobj_id=self.metadata["qobj_id"],
            job_id=self.job_id(),
            success=True,
            results=experiment_results,
        )

    def _get_job_results(self) -> Response:
        """Retrieves the results of this job from the backend"""
        backend: "TergiteBackend" = self.backend()
        provider: "Provider" = backend.provider
        url = f"{backend.base_url}{REST_API_MAP['jobs']}/{self.job_id()}"
        auth_headers = provider.get_auth_headers()
        return requests.get(url, headers=auth_headers)

    def __repr__(self):
        kwargs = [f"{k}={repr(v)}" for k, v in self.__dict__.items()]
        kwargs_str = ",\n".join(kwargs)
        return f"{self.__class__.__name__}({kwargs_str})"

    def __eq__(self, other):
        if not isinstance(other, Job):
            return False

        for k, v in self.__dict__.items():
            if getattr(other, k, None) != v:
                return False

        return True



# This code is part of Tergite
#
# (C) Copyright Miroslav Dobsicek 2020, 2021
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


from .backend import OpenPulseBackend, OpenQASMBackend
from .factory import Factory
from .job import Job
from .provider import Provider
from .version import __version__

Tergite = Factory()



# This code is part of Tergite
#
# (C) Copyright Axel Andersson 2022
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# This code was refactored from the original on 22nd September, 2023 by Martin Ahindura
"""Handles the calibration of the devices of type OpenPulseBackend"""
from dataclasses import dataclass
from typing import TYPE_CHECKING

import qiskit.circuit as circuit
from qiskit.transpiler import InstructionProperties, Target

from . import template_schedules as templates

if TYPE_CHECKING:
    from .backend import OpenPulseBackend


# TODO: Replace with BCC graph node
@dataclass(frozen=True)
class Node:
    """Parsing node for job objects from the API

    Note: the keys in the job objects should be
    valid Python string identifiers

    This node transforms the keys in the job object
    into attributes

    Attributes:
        data: the data to instantiate the node with
    """

    data: dict

    def __post_init__(self):
        # subvert frozen restriction, create new frozen fields for key value pairs in data
        for k, v in self.data.items():
            object.__setattr__(self, k, v)

    def __repr__(self) -> str:
        return self.job_id


def add_instructions(*, backend: "OpenPulseBackend", qubits: tuple, target: Target):
    """Adds Rx, Rz gates and Delay and Measure instructions for each qubit
    to the transpiler target provided.

    Implementations are backend dependent and are based on a specific calibration.

    Args:
        backend: the instance of
            :class:`~tergite.providers.tergite.backend:TergiteBackend`
            for which to add the instructions
        qubits: a tuple of qubits on which the instruction is to be run
        target: the target on which the instruction is to be added
    """
    # TODO: Fetch error statistics of gates from database

    rx_theta = circuit.Parameter("theta")
    rz_lambda = circuit.Parameter("lambda")
    delay_tau = circuit.Parameter("tau")

    # Reset qubits to ground state
    reset_props = {
        (q,): InstructionProperties(
            error=0.0,
            calibration=templates.delay(backend, (q,), 300 * 1000, delay_str="Reset"),
        )
        for q in qubits
    }
    target.add_instruction(circuit.library.Reset(), reset_props)

    # Rotation around X-axis on Bloch sphere
    rx_props = {
        (q,): InstructionProperties(
            error=0.0, calibration=templates.rx(backend, (q,), rx_theta)
        )
        for q in qubits
    }
    target.add_instruction(circuit.library.standard_gates.RXGate(rx_theta), rx_props)

    # Rotation around Z-axis on Bloch sphere
    rz_props = {
        (q,): InstructionProperties(
            error=0.0, calibration=templates.rz(backend, (q,), rz_lambda)
        )
        for q in qubits
    }
    target.add_instruction(circuit.library.standard_gates.RZGate(rz_lambda), rz_props)

    # Delay instruction
    delay_props = {
        (q,): InstructionProperties(
            error=0.0, calibration=templates.delay(backend, (q,), delay_tau)
        )
        for q in qubits
    }
    target.add_instruction(circuit.Delay(delay_tau), delay_props)

    # Measurement
    measure_props = {
        qubits: InstructionProperties(
            error=0.0, calibration=templates.measure(backend, qubits)
        )
    }
    target.add_instruction(circuit.measure.Measure(), measure_props)



# This code is part of Tergite
#
# (C) Copyright Miroslav Dobsicek 2020, 2021
# (C) Copyright Axel Andersson 2022
# (C) Copyright Martin Ahindura 2023
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# This code was refactored from the original on 22nd September, 2023 by Martin Ahindura
"""Defines the backend types provided by Tergite."""
import dataclasses
import functools
from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

import qiskit.circuit as circuit
import qiskit.compiler as compiler
import qiskit.pulse as pulse
import requests
from numpy import inf as infinity
from qiskit.circuit import QuantumCircuit
from qiskit.providers import BackendV2, Options, Provider
from qiskit.providers.models import BackendConfiguration
from qiskit.pulse.channels import (
    AcquireChannel,
    ControlChannel,
    DriveChannel,
    MeasureChannel,
    MemorySlot,
)
from qiskit.qobj import PulseQobj, QasmQobj
from qiskit.transpiler import Target
from qiskit.transpiler.coupling import CouplingMap

from tergite.qiskit.providers import calibrations

from .config import REST_API_MAP
from .job import Job

if TYPE_CHECKING:
    from .provider import Provider as TergiteProvider


class TergiteBackend(BackendV2):
    """Abstract class for Tergite Backends"""

    max_shots = infinity
    max_circuits = infinity

    def __init__(
        self, /, *, data: "TergiteBackendConfig", provider: Provider, base_url: str
    ):
        """Initialize a TergiteBackend based backend

        Args:
            provider: An optional backwards reference to the
                :class:`~qiskit.providers.Provider` object that the backend
                is from
            data: An optional config of type :class:`TergiteBackendConfig`
                from which to construct this backend
            base_url: An optional URL to Tergite API through which
                jobs are to be run

        Raises:
            AttributeError: If a field is specified that's outside the backend's
                options
            TypeError: If ``data`` is not a :class:`TergiteBackendConfig` object
        """
        super().__init__(
            provider=provider,
            name=data.name,
            backend_version=data.version,
        )
        self.base_url = base_url
        self.data = dataclasses.asdict(data)

    def register_job(self) -> Job:
        """Registers a new asynchronous job with the Tergite API.

        Returns:
             tergite.qiskit.providers.job.Job: An asynchronous
                 job registered in the Tergite API to be executed
        """
        jobs_url = self.base_url + REST_API_MAP["jobs"]
        provider: "TergiteProvider" = self.provider
        auth_headers = provider.get_auth_headers()
        response = requests.post(
            jobs_url,
            headers=auth_headers,
            params=dict(backend=self.name),
        )
        if response.ok:
            job_registration = response.json()
        else:
            raise RuntimeError(
                f"Unable to register job at the Tergite MSS, response: {response}"
            )
        job_id = job_registration["job_id"]
        job_upload_url = job_registration["upload_url"]
        return Job(backend=self, job_id=job_id, upload_url=job_upload_url)

    def run(self, experiments, /, **kwargs) -> Job:
        """Run on the Tergite backend.

        This method returns a :class:`~qiskit.providers.Job` object
        that runs circuits. Depending on the backend this may be either an async
        or sync call. It is at the discretion of the provider to decide whether
        running should block until the execution is finished or not: the Job
        class can handle either situation.

        Args:
            experiments (QuantumCircuit or Schedule or ScheduleBlock or list): An
                individual or a list of
                :class:`~qiskit.circuits.QuantumCircuit,
                :class:`~qiskit.pulse.ScheduleBlock`, or
                :class:`~qiskit.pulse.Schedule` objects to run on the backend.
            kwargs: Any kwarg options to pass to the backend for running the
                config. If a key is also present in the options
                attribute/object then the expectation is that the value
                specified will be used instead of what's set in the options
                object.

        Returns:
            tergite.qiskit.providers.job.Job: The job object for the run
        """
        job = self.register_job()
        qobj = self.make_qobj(experiments, **kwargs)
        response = job.submit(qobj)
        if response.ok:
            print("Tergite: Job has been successfully submitted")
            return job
        else:
            raise RuntimeError(
                f"Unable to transmit job to the Tergite BCC, response: {response}"
            )

    @abstractmethod
    def configuration(self) -> BackendConfiguration:
        """Retrieves this backend's configuration

        Returns:
            qiskit.providers.models.backendconfiguration.BackendConfiguration:
                this backend's configuration
        """
        ...

    @abstractmethod
    def make_qobj(self, experiments: list, /, **kwargs) -> object:
        """Constructs a Qobj from a list of user OpenPulse schedules or OpenQASM circuits

        Args:
            experiments (QuantumCircuit or Schedule or ScheduleBlock or list): An
                individual or a list of
                :class:`~qiskit.circuits.QuantumCircuit,
                :class:`~qiskit.pulse.ScheduleBlock`, or
                :class:`~qiskit.pulse.Schedule` objects to run on the backend.

        Returns:
             qiskit.qobj.pulse_qobj.PulseQobj or qiskit.qobj.pulse_qobj.QasmQobj
                transpiled from the experiments.
        """
        ...

    @property
    def meas_map(self) -> list:
        return self.data["meas_map"]

    @functools.cached_property
    def coupling_map(self) -> CouplingMap:
        # It is required that nodes are contiguously indexed starting at 0.
        # Missed nodes will be added as isolated nodes in the coupling map.
        #
        # Also for some reason the graph has to be connected in Qiskit? Not sure why that is..
        edges = sorted(
            sorted(self.data["coupling_map"], key=lambda i: i[1]), key=lambda i: i[0]
        )
        return CouplingMap(couplinglist=list(edges))

    @classmethod
    def _default_options(cls, /) -> Options:
        """This defines the default user configurable settings of this backend.

        See: help(backend.set_options)

         Returns:
            qiskit.providers.Options: A options object with
                default values set
        """
        options = Options(shots=2000)
        options.set_validator("shots", (1, TergiteBackend.max_shots))
        return options

    def _as_dict(self):
        """Converts this backend into a dictionary representation

        Returns:
            dict: a dictionary representation of this backend
        """
        obj = self.configuration().to_dict()
        obj["characterized"] = self.data["characterized"]
        return obj

    def __repr__(self) -> str:
        repr_list = [f"TergiteBackend object @ {hex(id(self))}:"]
        config = self._as_dict()
        for attr, value in config.items():
            repr_list.append(f"  {attr}:\t{value}".expandtabs(30))
        return "\n".join(repr_list)

    def __eq__(self, other: Any):
        if not isinstance(other, TergiteBackend):
            return False

        self_dict = self._as_dict().copy()
        other_dict = other._as_dict().copy()

        # serialize a few items that are hard to serialize
        self_dict["coupling_map"] = f"{self_dict['coupling_map']}"
        other_dict["coupling_map"] = f"{other_dict['coupling_map']}"

        self_dict["supported_instructions"] = f"{self_dict['supported_instructions']}"
        other_dict["supported_instructions"] = f"{other_dict['supported_instructions']}"

        return self_dict == other_dict


class OpenPulseBackend(TergiteBackend):
    open_pulse = True

    parametric_pulses = [
        "constant",
        "zero",
        "square",
        "sawtooth",
        "triangle",
        "cos",
        "sin",
        "gaussian",
        "gaussian_deriv",
        "sech",
        "sech_deriv",
        "gaussian_square",
        "drag",
    ]

    def configuration(self) -> BackendConfiguration:
        return BackendConfiguration(
            backend_name=self.name,  # From BackendV2.
            backend_version=self.backend_version,  # From BackendV2.
            n_qubits=self.target.num_qubits,  # Number of qubits, obtained from self.target.
            basis_gates=[],  # Don't need.
            gates=[],  # Don't need.
            simulator=False,  # This is a real quantum computer.
            conditional=False,  # We cannot do conditional gate application yet.
            local=False,  # Jobs are sent over the internet.
            open_pulse=True,  # This backend converts all quantum experiment input to schedules.
            meas_levels=(0, 1, 2),  # 0: RAW, 1: INTEGRATED, 2: DISCRIMINATED
            memory=False,  # Unsure what this is for. (?)
            max_shots=TergiteBackend.max_shots,  # From TergiteBackend.
            coupling_map=self.coupling_map,  # From TergiteBackend.
            supported_instructions=self.target.instructions,  # Supported instructions, obtained from self.target.
            dt=self.dt,  # From self
            dtm=self.dtm,  # From self
            description=self.description,  # From BackendV2.
            parametric_pulses=OpenPulseBackend.parametric_pulses,  # From type of self
        )

    @property
    def dt(self) -> float:
        return self.data["dt"]

    @property
    def dtm(self) -> float:
        return self.data["dtm"]

    @functools.cached_property
    def target(self) -> Target:
        gmap = Target(num_qubits=self.data["num_qubits"], dt=self.data["dt"])
        if self.data["characterized"]:
            calibrations.add_instructions(
                backend=self,
                qubits=tuple(q for q in range(self.data["num_qubits"])),
                target=gmap,
            )
        return gmap

    @functools.cached_property
    def device_properties(self) -> dict:
        """the collection of properties specific to this backend"""
        return self.data["device_properties"]

    @property
    def qubit_lo_freq(self) -> list:
        return [0.0] * self.data["num_qubits"]

    @property
    def meas_lo_freq(self) -> list:
        return [0.0] * self.data["num_resonators"]

    def drive_channel(self, qubit_idx: int) -> DriveChannel:
        return DriveChannel(qubit_idx)

    def measure_channel(self, qubit_idx: int) -> MeasureChannel:
        return MeasureChannel(qubit_idx)

    def acquire_channel(self, qubit_idx: int) -> AcquireChannel:
        return AcquireChannel(qubit_idx)

    def memory_slot(self, qubit_idx: int) -> MemorySlot:
        return MemorySlot(qubit_idx)

    def control_channel(self, qubits: list) -> list:
        if qubits not in self.coupling_map.get_edges():
            raise ValueError(f"Coupling {qubits} not in coupling map.")

        return list(map(ControlChannel, qubits))

    def make_qobj(self, experiments: object, /, **kwargs) -> PulseQobj:
        if type(experiments) is not list:
            experiments = [experiments]

        # convert all non-schedules to schedules
        experiments = [
            compiler.schedule(experiment, backend=self)
            if (type(experiment) is not pulse.ScheduleBlock)
            and (type(experiment) is not pulse.Schedule)
            else experiment  # already a schedule, so don't convert
            for experiment in experiments
        ]

        # assemble schedules to PulseQobj
        return compiler.assemble(
            experiments=experiments,
            backend=self,
            shots=self.options.shots,
            qubit_lo_freq=self.qubit_lo_freq,
            meas_lo_freq=self.meas_lo_freq,
            **kwargs,
        )


class OpenQASMBackend(TergiteBackend):
    open_pulse = False

    @functools.cached_property
    def target(self) -> Target:
        gmap = Target(num_qubits=self.data["num_qubits"])

        # add rx, rz, delay, measure gate set
        gmap.add_instruction(circuit.library.Reset())
        gmap.add_instruction(
            circuit.library.standard_gates.RXGate(circuit.Parameter("theta"))
        )
        gmap.add_instruction(
            circuit.library.standard_gates.RZGate(circuit.Parameter("lambda"))
        )
        gmap.add_instruction(circuit.Delay(circuit.Parameter("tau")))
        gmap.add_instruction(circuit.measure.Measure())

        return gmap

    def make_qobj(
        self, experiments: Union[List[QuantumCircuit], QuantumCircuit], /, **kwargs
    ) -> QasmQobj:
        """Constructs a QasmQobj from an OpenQASM circuit, or a list of them

        Raises:
            TypeError: If any of the experiments passed in not an instance
                of :class:`~qiskit.qobj.qasm_qobj.QasmQobj`

        Returns:
            qiskit.qobj.qasm_qobj.QasmQobj: A QasmQobj object transpiled from the circuits
        """
        if type(experiments) is not list:
            experiments = [experiments]
        for e in experiments:
            if not isinstance(e, QuantumCircuit):
                raise TypeError(f"Experiment {e} is not an instance of QuantumCircuit.")

        circuits = compiler.transpile(circuits=experiments, backend=self)
        return compiler.assemble(
            experiments=circuits, shots=self.options.shots, **kwargs
        )

    def configuration(self) -> BackendConfiguration:
        return BackendConfiguration(
            backend_name=self.name,  # From BackendV2.
            backend_version=self.backend_version,  # From BackendV2.
            n_qubits=self.target.num_qubits,  # Number of qubits, obtained from self.target.
            basis_gates=[],  # Don't need.
            gates=[],  # Don't need.
            simulator=False,  # This is a real quantum computer.
            conditional=False,  # We cannot do conditional gate application yet.
            local=False,  # Jobs are sent over the internet.
            open_pulse=False,  # This backend just transpiles circuits.
            meas_levels=(2,),  # 0: RAW, 1: INTEGRATED, 2: DISCRIMINATED
            memory=True,  # Meas level 2 results are stored in MongoDB (?).
            max_shots=TergiteBackend.max_shots,  # From TergiteBackend.
            coupling_map=self.coupling_map,  # From TergiteBackend.
        )


@dataclasses.dataclass
class TergiteBackendConfig:
    """Basic structure of the config of a backend"""

    name: str
    characterized: bool
    open_pulse: bool
    timelog: Dict[str, Any]
    version: str
    meas_map: List[List[int]]
    coupling_map: List[Tuple[int, int]]
    description: str = None
    simulator: bool = False
    num_qubits: int = 0
    num_couplers: int = 0
    num_resonators: int = 0
    online_date: Optional[str] = None
    dt: Optional[float] = None
    dtm: Optional[float] = None
    qubit_ids: Dict[int, str] = dataclasses.field(default_factory=dict)
    device_properties: Optional["_DeviceProperties"] = None
    discriminators: Optional[Dict[str, Any]] = None
    meas_lo_freq: Optional[List[int]] = None
    qubit_lo_freq: Optional[List[int]] = None
    qubit_calibrations: Optional[Dict[str, Any]] = None
    coupler_calibrations: Optional[Dict[str, Any]] = None
    resonator_calibrations: Optional[Dict[str, Any]] = None
    gates: Optional[Dict[str, Any]] = None
    is_active: Optional[bool] = None
    # TODO: Remove immediately, just for testing
    properties: Optional[Dict[str, Any]] = None

    def __post_init__(self):
        """Run after initialization of the dataclass"""
        # convert nested dataclasses to dataclasses
        if isinstance(self.device_properties, dict):
            self.device_properties = _DeviceProperties(**self.device_properties)


@dataclasses.dataclass
class _DeviceProperties:
    """All Device Properties"""

    qubit: Optional[List["_QubitProps"]] = None
    readout_resonator: Optional[List["_ReadoutResonatorProps"]] = None
    coupler: Optional[List[Dict[str, Any]]] = None

    def __post_init__(self):
        """Run after initialization of the dataclass"""
        # convert nested dataclasses to dataclasses
        if isinstance(self.qubit, list):
            self.qubit = [
                _QubitProps(**item) for item in self.qubit if isinstance(item, dict)
            ]

        if isinstance(self.readout_resonator, list):
            self.readout_resonator = [
                _ReadoutResonatorProps(**item)
                for item in self.readout_resonator
                if isinstance(item, dict)
            ]


@dataclasses.dataclass
class _ReadoutResonatorProps:
    """ReadoutResonator Device configuration"""

    acq_delay: float
    acq_integration_time: float
    frequency: int
    pulse_amplitude: float
    pulse_delay: float
    pulse_duration: float
    pulse_type: str
    id: Optional[int] = None
    index: Optional[int] = None
    x_position: Optional[int] = None
    y_position: Optional[int] = None
    readout_line: Optional[int] = None
    lda_parameters: Optional[Dict[str, Any]] = None


@dataclasses.dataclass
class _QubitProps:
    """Qubit Device configuration"""

    frequency: int
    pi_pulse_amplitude: float
    pi_pulse_duration: float
    pulse_type: str
    pulse_sigma: float
    t1_decoherence: float
    t2_decoherence: float
    id: Optional[int] = None
    index: Optional[int] = None
    x_position: Optional[int] = None
    y_position: Optional[int] = None
    xy_drive_line: Optional[int] = None
    z_drive_line: Optional[int] = None



# This code is part of Tergite
#
# (C) Copyright Axel Andersson 2022
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# This code was modified from the original by:
#
# - Martin Ahindura 2023
# - Stefan Hill 2023
"""Handles the creation of schedules for the devices of type OpenPulseBackend"""
from typing import TYPE_CHECKING, Iterable

import numpy as np
import qiskit.circuit as circuit
import qiskit.pulse as pulse

if TYPE_CHECKING:
    from .backend import OpenPulseBackend


def rx(
    backend: "OpenPulseBackend", qubits: Iterable, rx_theta: circuit.Parameter
) -> pulse.ScheduleBlock:
    """Creates a rotation around the x-axis on the Bloch sphere for the list of qubits in the backend.

    It returns a backend-specific schedule which implements a
    rotation on a set of qubits around the x-axis on the Bloch sphere.
    The schedule is parameterised by θ [rad], which specifies the angle.

    Args:
        backend: the backend for which the schedule is to be created.
        qubits: the set of qubits to be affected.
        rx_theta: the theta parameter for the circuit

    Returns:
        qiskit.pulse.ScheduleBlock: the schedule implementing the rotation
    """
    device_properties = backend.device_properties
    qubit = device_properties.get("qubit")

    sched = pulse.ScheduleBlock(name=f"RX(θ, {qubits})")
    for q in qubits:
        sched += pulse.SetFrequency(
            qubit[q]["frequency"],
            channel=backend.drive_channel(q),
        )
        sched += pulse.Play(
            pulse.Gaussian(
                duration=round(qubit[q].get("pi_pulse_duration") / backend.dt),
                amp=rx_theta / np.pi * qubit[q].get("pi_pulse_amplitude"),
                sigma=round(qubit[q].get("pulse_sigma") / backend.dt),
                name=f"RX q{q}",
            ),
            channel=backend.drive_channel(q),
        )

    return sched


def rz(
    backend: "OpenPulseBackend", qubits: Iterable, rz_lambda: circuit.Parameter
) -> pulse.ScheduleBlock:
    """Creates a rotation around the z-axis on the Bloch sphere for the list of qubits in the backend.

    It returns a backend-specific schedule which implements a rotation on a set
    of qubits around the z-axis on the Block sphere. The schedule is
    parameterised by λ [rad], which specifies the angle.

    Args:
        backend: the backend for which the schedule is to be created.
        qubits: the set of qubits to be affected.
        rz_lambda: the lambda parameter for the circuit

    Returns:
        qiskit.pulse.ScheduleBlock: the schedule implementing the rotation
    """
    sched = pulse.ScheduleBlock(name=f"RZ(λ, {qubits})")
    for q in qubits:
        sched += pulse.ShiftPhase(
            rz_lambda, channel=backend.drive_channel(q), name=f"RZ q{q}"
        )
    return sched


def measure(backend: "OpenPulseBackend", qubits: Iterable) -> pulse.ScheduleBlock:
    """Creates a measurement on the list of qubits in the given backend.

    It returns a backend-specific schedule which implements
    a measurement on a set of qubits.

    Args:
        backend: the backend for which the schedule is to be created.
        qubits: the set of qubits to be affected.

    Returns:
        qiskit.pulse.ScheduleBlock: the schedule implementing the measurement
    """
    device_properties = backend.device_properties
    readout_resonator_props = device_properties.get("readout_resonator")

    sched = pulse.ScheduleBlock(name=f"Measure({qubits})")
    for q in qubits:
        readout_resonator = readout_resonator_props[q]
        sched += pulse.SetFrequency(
            readout_resonator.get("frequency"),
            channel=backend.measure_channel(q),
        )
        sched += pulse.Play(
            pulse.Constant(
                amp=readout_resonator.get("pulse_amplitude"),
                duration=round(readout_resonator.get("pulse_duration") / backend.dt),
                name=f"Readout q{q}",
            ),
            channel=backend.measure_channel(q),
        )
        sched += pulse.Delay(
            duration=300,
            channel=backend.acquire_channel(q),
            name=f"Time of flight q{q}",
        )
        sched += pulse.Acquire(
            duration=round(readout_resonator.get("acq_integration_time") / backend.dt),
            channel=backend.acquire_channel(q),
            mem_slot=backend.memory_slot(q),
            name=f"Integration window q{q}",
        )
    return sched


def delay(
    backend: "OpenPulseBackend",
    qubits: Iterable,
    delay_tau: circuit.Parameter,
    delay_str: str = "Delay",
) -> pulse.ScheduleBlock:
    """Creates a delay on the set of qubits in the backend.

    It returns a backend-specific schedule which implements a delay
    operation on a set of qubits.
    The schedule is parameterised by τ [ns], which specifies the delay duration.

    Args:
        backend: the backend for which the schedule is to be created.
        qubits: the set of qubits to be affected.
        delay_tau: the tau parameter for the circuit
        delay_str: the delay string for the instruction set

    Returns:
        qiskit.pulse.ScheduleBlock: the schedule implementing the delay
    """
    sched = pulse.ScheduleBlock(name=f"{delay_str}({qubits}, τ)")
    for q in qubits:
        sched += pulse.Delay(
            duration=delay_tau,
            channel=backend.drive_channel(q),
            name=f"{delay_str} q{q}",
        )
        sched += pulse.Delay(
            duration=delay_tau,
            channel=backend.measure_channel(q),
            name=f"{delay_str} q{q}",
        )
        sched += pulse.Delay(
            duration=delay_tau,
            channel=backend.acquire_channel(q),
            name=f"{delay_str} q{q}",
        )
    return sched
